import pytest
import talonfmt


@pytest.mark.golden_test("golden/default/*.yml")
def test_golden_default(golden):
    output = talonfmt.talonfmt(
        contents=golden["input"]
    )
    assert output == golden.out["output"]


# @pytest.mark.golden_test("golden/align/dynamic/*.yml")
# def test_golden_align_dynamic(golden):
#     output = talonfmt.talonfmt(
#         contents=golden["input"],
#         align_match_context=True,
#         align_short_commands=True
#     )
#     assert output == golden.out["output"]


# @pytest.mark.golden_test("golden/align/fixed32/*.yml")
# def test_golden_align_fixed32(golden):
#     output = talonfmt.talonfmt(
#         contents=golden["input"],
#         align_match_context=True,
#         align_match_context_at=32,
#         align_short_commands=True,
#         align_short_commands_at=32,
#     )
#     assert output == golden.out["output"]



from doc_printer import SmartDocRenderer, TokenStream
from pathlib import Path
from talonfmt.formatter import TalonFormatter
from typing import Iterator, Optional, TextIO, Union

import click
import io
import sys
import tokenize
import tree_sitter_talon


def talonfmt(
    contents: str,
    *,
    encoding: str = "utf-8",
    indent_size: int = 4,
    max_line_width: int = 80,
    align_match_context: bool = False,
    align_match_context_at: Optional[int] = None,
    align_short_commands: bool = False,
    align_short_commands_at: Optional[int] = None,
) -> str:
    # Enable align_match_context if align_match_context_at is set:
    merged_match_context: Union[bool, int]
    if isinstance(align_match_context_at, int):
        merged_match_context = align_match_context_at
    else:
        merged_match_context = align_match_context

    # Enable align_short_commands if align_short_commands_at is set:
    merged_short_commands: Union[bool, int]
    if isinstance(align_short_commands_at, int):
        merged_short_commands = align_short_commands_at
    else:
        merged_short_commands = align_short_commands

    # Create an instance of TalonFormatter
    talon_formatter = TalonFormatter(
        indent_size=indent_size,
        align_match_context=merged_match_context,
        align_short_commands=merged_short_commands,
    )

    # Create an instance of DocRenderer
    doc_renderer = SmartDocRenderer(
        max_line_width=max_line_width,
    )

    ast = tree_sitter_talon.parse(contents, encoding=encoding)
    doc = talon_formatter.format(ast)
    return doc_renderer.to_str(doc)


@click.command(name="talonfmt")
@click.argument(
    "input",
    required=False,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=True, readable=True, path_type=Path
    ),
)
@click.option("--indent-size", type=int, default=4, show_default=True)
@click.option(
    "--align-match-context/--no-align-match-context",
    default=False,
    show_default=True,
)
@click.option(
    "--align-match-context-at",
    type=int,
)
@click.option(
    "--align-short-commands/--no-align-short-commands",
    default=False,
    show_default=True,
)
@click.option(
    "--align-short-commands-at",
    type=int,
)
@click.option(
    "--in-place/--stdout",
    default=False,
    show_default=True,
)
@click.option("--max-line-width", type=int, default=80, show_default=True)
def cli(
    *,
    input: Optional[Path],
    indent_size: int,
    max_line_width: int,
    align_match_context: bool,
    align_match_context_at: Optional[int],
    align_short_commands: bool,
    align_short_commands_at: Optional[int],
    in_place: bool,
):
    def readfile(filename: Path) -> tuple[str, str]:
        with filename.open(mode="rb") as fp:
            bytes_on_disk = fp.read()
        encoding, _ = tokenize.detect_encoding(io.BytesIO(bytes_on_disk).readline)
        with io.TextIOWrapper(io.BytesIO(bytes_on_disk), encoding) as wrapper:
            contents = wrapper.read()
        return (contents, encoding)

    def format(contents: str, *, encoding: str) -> str:
        return talonfmt(
            contents,
            encoding=encoding,
            indent_size=indent_size,
            max_line_width=max_line_width,
            align_match_context=align_match_context,
            align_match_context_at=align_match_context_at,
            align_short_commands=align_short_commands,
            align_short_commands_at=align_short_commands_at,
        )

    def format_file(filename: Path):
        contents, encoding = readfile(filename)
        output = format(contents, encoding=encoding)
        if in_place:
            with filename.open(mode="w") as handle:
                handle.write(output)
        else:
            sys.stdout.write(output)

    if input:
        if input.is_file():
            format_file(input)
        if input.is_dir():
            for filename in input.glob("**/*.talon"):
                format_file(filename)
    else:
        contents = "\n".join(sys.stdin.readlines())
        encoding = sys.stdin.encoding
        sys.stdout.write(format(contents, encoding=encoding))


# with filename.open(mode="w", encoding=encoding) as fp:
def main():
    cli(prog_name="talonfmt")


if __name__ == "__main__":
    main()



from collections.abc import Iterator, Sequence
from functools import singledispatchmethod
from typing import Iterable, Union
from doc_printer import *
from tree_sitter_talon import *

import dataclasses


class FormatError(Exception):
    pass


@dataclasses.dataclass
class TalonFormatter:
    indent_size: int
    align_match_context: Union[bool, int]
    align_short_commands: Union[bool, int]

    @singledispatchmethod
    def format(self, node: Node) -> Doc:
        raise TypeError(type(node))

    def format_children(self, children: Iterable[Node]) -> Iterator[Doc]:
        for child in self.store_comments(children):
            if isinstance(child, Iterable):
                yield from self.format_children(child)
            else:
                yield self.format(child)

    @format.register
    def _(self, node: TalonError) -> Doc:
        raise ValueError(node.start_position, node.end_position)

    ###########################################################################
    # Documents
    ###########################################################################

    @format.register
    def _(self, node: TalonSourceFile) -> Doc:
        header: list[Doc] = []
        body: list[Doc] = []
        for child in self.store_comments(node.children):
            if isinstance(child, TalonContext):
                header.extend(self.get_comments())
                context = self.format(child)
                if context:
                    header.append(context)
            else:
                body.extend(self.get_comments())
                body.append(self.format(child))
        body_with_tables = create_tables(iter(body))
        return Line.join(header, "-", body_with_tables)

    ###########################################################################
    # Tag Includes
    ###########################################################################

    @format.register
    def _(self, node: TalonIncludeTag) -> Doc:
        # NOTE: top-level constructs should not include a trailing newline
        self.only_comment_children(node)
        include_tag = "tag():" // self.format(node.tag)
        return Line.join(self.get_comments(), include_tag)

    ###########################################################################
    # Settings
    ###########################################################################

    @format.register
    def _(self, node: TalonSettings) -> Doc:
        # NOTE: top-level constructs should not include a trailing newline
        settings = cat(
            "settings():",
            nest(self.indent_size, Line, self.format(self.get_only_child(node))),
        )
        return Line.join(self.get_comments(), settings)

    ###########################################################################
    # Commands
    ###########################################################################

    @format.register
    def _(self, node: TalonCommand) -> Doc:
        # NOTE: top-level constructs should not include a trailing newline

        rule_doc = self.format(node.rule)
        rule_comments = list(self.get_comments())
        script_doc = self.format(node.script)
        script_comments = list(self.get_comments())
        # (1): a line-break after the rule, e.g.,
        #
        # select camel left:
        #     user.extend_camel_left()
        #
        alt1 = cat(
            rule_doc / ":",
            nest(
                self.indent_size,
                Line,
                Line.join(script_comments, script_doc),
            ),
        )

        # (2): the rule and a single-line talon script on the same line, e.g.,
        #
        # select camel left: user.extend_camel_left()
        #
        alt2 = self.aligned_command(
            rule_doc / ":",
            Line.join(script_comments, script_doc),
            multiline=len(script_comments) + len(node.script.children) > 1,
        )

        alts = alt1 | alt2
        return Line.join(rule_comments, alts) if rule_comments else alts

    def aligned_command(self, *doclike: DocLike, multiline: bool) -> Doc:
        if self.align_short_commands and not multiline:
            if self.align_short_commands is True:
                return row(*doclike, table_type="command")
            else:
                return row(
                    *doclike,
                    table_type="command",
                    min_col_widths=(self.align_short_commands,),
                )
        else:
            return Fail

    ###########################################################################
    # Statements and Blocks
    ###########################################################################

    @format.register
    def _(self, node: TalonBlock) -> Doc:
        docs: list[Doc] = []
        for stmt in node.children:
            doc = self.format(stmt)
            docs.extend(self.get_comments())
            docs.append(doc)
        return Line.join(docs)

    @format.register
    def _(self, node: TalonAssignment) -> Doc:
        self.only_comment_children(node)
        left = self.format(node.left)
        right = self.format(node.right)
        return Space.join(left, "=", right)

    @format.register
    def _(self, node: TalonComment) -> Doc:
        comment = node.text.strip().lstrip("#").lstrip()
        return "#" // Text.words(comment)

    @format.register
    def _(self, node: TalonDocstring) -> Doc:
        comment = node.text.strip().lstrip("#").lstrip()
        return "###" // Text.words(comment)

    @format.register
    def _(self, node: TalonExpression) -> Doc:
        self.only_comment_children(node)
        expression = self.format(node.expression)
        return expression

    ###########################################################################
    # Expressions
    ###########################################################################

    @format.register
    def _(self, node: TalonAction) -> Doc:
        self.only_comment_children(node)
        action_name = self.format(node.action_name)
        arguments = self.format(node.arguments)
        return action_name / parens(arguments)

    @format.register
    def _(self, node: TalonArgumentList) -> Doc:
        separator = "," / Space
        children = self.format_children(node.children)
        return separator.join(children)

    @format.register
    def _(self, node: TalonBinaryOperator) -> Doc:
        self.only_comment_children(node)
        left = self.format(node.left)
        operator = self.format(node.operator)
        right = self.format(node.right)
        return Space.join(left, operator, right)

    @format.register
    def _(self, node: TalonFloat) -> Doc:
        return Text.words(node.text)

    @format.register
    def _(self, node: TalonIdentifier) -> Doc:
        return Text.words(node.text)

    @format.register
    def _(self, node: TalonImplicitString) -> Doc:
        return Text.words(node.text)

    @format.register
    def _(self, node: TalonInteger) -> Doc:
        return Text.words(node.text)

    @format.register
    def _(self, node: TalonInterpolation) -> Doc:
        return self.format(self.get_only_child(node))

    @format.register
    def _(self, node: TalonKeyAction) -> Doc:
        self.only_comment_children(node)
        return "key" / parens(self.format(node.arguments))

    @format.register
    def _(self, node: TalonNumber) -> Doc:
        return self.format(self.get_only_child(node))

    @format.register
    def _(self, node: TalonOperator) -> Doc:
        return Text.words(node.text)

    @format.register
    def _(self, node: TalonParenthesizedExpression) -> Doc:
        return parens(self.format(self.get_only_child(node)))

    @format.register
    def _(self, node: TalonRegexEscapeSequence) -> Doc:
        if node.children:
            return braces(self.format_children(node.children))
        else:
            return braces(Empty)

    @format.register
    def _(self, node: TalonSleepAction) -> Doc:
        self.only_comment_children(node)
        return "sleep" / parens(self.format(node.arguments))

    @format.register
    def _(self, node: TalonString) -> Doc:
        return quote(self.format_children(node.children))

    @format.register
    def _(self, node: TalonStringContent) -> Doc:
        return Text.words(node.text)

    @format.register
    def _(self, node: TalonStringEscapeSequence) -> Doc:
        return Text.words(node.text)

    @format.register
    def _(self, node: TalonVariable) -> Doc:
        self.only_comment_children(node)
        return self.format(node.variable_name)

    ###########################################################################
    # Rules
    ###########################################################################

    @format.register
    def _(self, node: TalonCapture) -> Doc:
        self.only_comment_children(node)
        return angles(self.format(node.capture_name))

    @format.register
    def _(self, node: TalonChoice) -> Doc:
        children = self.format_children(node.children)
        operator = Space / "|" / Space
        return operator.join(children)

    @format.register
    def _(self, node: TalonEndAnchor) -> Doc:
        return Text("$")

    @format.register
    def _(self, node: TalonList) -> Doc:
        self.only_comment_children(node)
        return braces(self.format(node.list_name))

    @format.register
    def _(self, node: TalonOptional) -> Doc:
        child = self.get_only_child(node)
        return brackets(self.format(child))

    @format.register
    def _(self, node: TalonParenthesizedRule) -> Doc:
        child = self.get_only_child(node)
        return parens(self.format(child))

    @format.register
    def _(self, node: TalonRepeat) -> Doc:
        child = self.get_only_child(node)
        return self.format(child) / "*"

    @format.register
    def _(self, node: TalonRepeat1) -> Doc:
        return self.format(self.get_only_child(node)) / "+"

    @format.register
    def _(self, node: TalonRule) -> Doc:
        return cat(self.format_children(node.children))

    @format.register
    def _(self, node: TalonSeq) -> Doc:
        return Space.join(self.format_children(node.children))

    @format.register
    def _(self, node: TalonStartAnchor) -> Doc:
        return Text("^")

    @format.register
    def _(self, node: TalonWord) -> Doc:
        return Text.words(node.text)

    ###########################################################################
    # Comments
    ###########################################################################

    # Used to buffer comments encountered inline, e.g., inside a binary operator
    _comment_buffer: list[TalonComment] = dataclasses.field(
        default_factory=list, init=False
    )

    def store_comments(self, children: Iterable[Node]) -> Iterator[Node]:
        for child in children:
            if isinstance(child, TalonComment):
                self._comment_buffer.append(child)
            else:
                yield child

    def get_comments(self) -> Iterator[Doc]:
        try:
            for comment in self._comment_buffer:
                yield self.format(comment)
        finally:
            self._comment_buffer.clear()

    def only_comment_children(self, branch: Branch) -> None:
        if isinstance(branch.children, list):
            rest = tuple(self.store_comments(branch.children))
            assert (
                len(rest) == 0
            ), f"There should be no non-comment children, found {tuple(node.type_name for node in rest)} in {branch.type_name}:\n{branch}"

    def get_only_child(self, branch: Branch) -> Node:
        assert branch.children is not None
        if isinstance(branch.children, Sequence):
            rest = tuple(self.store_comments(branch.children))
            assert (
                len(rest) == 1
            ), f"There should be only one non-comment child, found {tuple(node.type_name for node in rest)} in {branch.type_name}:\n{branch}"
            return rest[0]
        else:
            return branch.children

    ###########################################################################
    # Context Header
    ###########################################################################

    @format.register
    def _(self, node: TalonContext) -> Doc:
        docstrings: list[Doc] = []
        buffer: list[Doc] = []
        for child in node.children:
            if isinstance(child, TalonDocstring):
                docstrings.append(self.format(child))
            elif isinstance(child, TalonComment):
                buffer.append(self.format(child))
            else:
                buffer.extend(self.format_matches(child, False, False))
        if not docstrings and not buffer:
            return Empty
        else:
            return Line.join(docstrings, buffer)

    @singledispatchmethod
    def format_matches(
        self,
        match: Union[TalonAnd, TalonNot, TalonMatch, TalonOr],
        *,
        under_and: bool,
        under_not: bool,
    ) -> Iterator[Doc]:
        raise TypeError(type(match))

    @format_matches.register
    def _(self, match: TalonAnd, under_and: bool, under_not: bool) -> Iterator[Doc]:
        for child in match.children:
            if isinstance(child, TalonComment):
                yield self.format(child)
            else:
                yield from self.format_matches(child, under_and, under_not)
                under_and = True

    @format_matches.register
    def _(self, match: TalonNot, under_and: bool, under_not: bool) -> Iterator[Doc]:
        for child in match.children:
            if isinstance(child, TalonComment):
                yield self.format(child)
            else:
                yield from self.format_matches(child, under_and, under_not)
                under_not = True

    @format_matches.register
    def _(self, match: TalonOr, under_and: bool, under_not: bool) -> Iterator[Doc]:
        for child in match.children:
            if isinstance(child, TalonComment):
                yield self.format(child)
            else:
                yield from self.format_matches(child, under_and, under_not)

    @format_matches.register
    def _(self, match: TalonMatch, under_and: bool, under_not: bool) -> Iterator[Doc]:
        kwds = self.match_keywords(under_and, under_not)
        key = cat(kwds, self.format(match.key), ":")
        pattern = self.format(match.pattern)
        yield alt(self.create_match_alts(key, pattern))

    def match_keywords(self, under_and: bool, under_not: bool) -> Iterator[Doc]:
        if under_and:
            yield Text("and")
            yield Space
        if under_not:
            yield Text("not")
            yield Space

    def create_match_alts(self, *doclike: DocLike) -> Iterator[Doc]:
        # Standard
        yield Space.join(doclike)

        # Aligned alternative
        if self.align_match_context:
            if self.align_match_context is True:
                yield row(*doclike, table_type="match")
            else:
                yield row(
                    *doclike,
                    table_type="match",
                    min_col_widths=(self.align_match_context,),
                )
